# ---------------------------------------------------------------------------- #
#                           Load packages and modules                          #
# ---------------------------------------------------------------------------- #

import pandas as pd
import numpy as np

# ----------------------------- Standard library ----------------------------- #

import os
import sys
from itertools import compress
from re import sub
from keyword import iskeyword
from collections import namedtuple
from typing import List, Dict, Tuple, Union, Optional, NamedTuple

# ------------------------------- Intra-package ------------------------------ #

from my_mltools.exceptions import (ColumnNameKeyWordError,
                                   ColumnNameStartWithDigitError,
                                   InvalidIdentifierError,
                                   InvalidColumnDtypeError)
from my_mltools.check_utils import is_list_str, is_string_dtype_pd

# ---------------------------------------------------------------------------- #
#                               Cleaning helpers                               #
# ---------------------------------------------------------------------------- #

# ------------------ Function to check column name integrity ----------------- #


def check_col_nms(df: pd.DataFrame) -> None:
    """
    This function is a helper that checks the integrity of the column names. 

    Parameters
    ----------
    df : DataFrame

    Raises
    ------
    TypeError
        The argument 'df' must be a DataFrame.
    ColumnNameKeyWordError
        Columns name cannot contain reserved keywords like 'def', 'for'.
    ColumnNameStartWithDigitError
        Column names cannot begin with a digit.
    InvalidIdentifierError
        Catch-all exception for invalid identifiers.
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError("'df' must be a DataFrame")

    col_nms = df.columns.tolist()

    # Check column names are not keywords
    kw_index = [iskeyword(col) for col in col_nms]
    if any(kw_index):
        raise ColumnNameKeyWordError(list(compress(col_nms, kw_index)))

    # Check column names do not start with digit
    digit_index = [col[0].isdigit() for col in col_nms]
    if any(digit_index):
        raise ColumnNameStartWithDigitError(
            list(compress(col_nms, digit_index)))

    # Catch-all check
    invalid_identifier_index = [not col.isidentifier() for col in col_nms]
    if any(invalid_identifier_index):
        raise InvalidIdentifierError(
            list(compress(col_nms, invalid_identifier_index)))


# --------------------- Function that cleans column names -------------------- #


def clean_col_nms(df: pd.DataFrame, inplace: Optional[bool] = False) -> Union[pd.DataFrame, None]:
    """
    This helper function removes any invalid character, e.g. special characters and white spaces, in a column name and removes 
    leading characters until a character from a-z or A-Z is matched. Note this function does not replace python keywords or reserved 
    words if they exist as column names. Use the `rename()` method of pandas DataFrame or the datatable rename `{"A": "col_A"}` syntax 
    to clean the column names if `check_col_nms()` reveals such invalid columns names. 

    Parameters
    ----------
    df : Dataframe
    inplace : bool, optional
        Whether to return a new DataFrame, by default False.

    Returns
    -------
    DataFrame
        A DataFrame with transformed column names or None if inplace=True.
    """
    # Create a copy if inplace=False
    if (not inplace):
        df = df.copy()

    original_col_nms = df.columns.tolist()
    # Remove leading characters until a letter is matched
    # ^ matches the beginning of the string
    # [^ ] is negated set, matching any character that is not in this set
    # + is a quantifier, matching the preceding element one or more times
    new_col_nms = (sub('^[^a-zA-Z]+', '', col) for col in original_col_nms)
    # Remove trailing characters until a letter is matched
    new_col_nms = (sub('[^a-zA-Z]+$', '', col) for col in new_col_nms)
    # Replace white spaces in-between words with "_"
    # \s+ matches 1 or more whitespace characters (spaces, tabs, line breaks)
    new_col_nms = (sub(r'\s+', '_', col) for col in new_col_nms)
    # [^a-zA-Z0-9_] matches any character that is not a 'word character' (alphanumeric & underscore), which is equivalent to \W
    df.columns = [sub('[^a-zA-Z0-9_]', '', col.lower()) for col in new_col_nms]

    # Return copy
    if (not inplace):
        return df
    else:
        return None


# ----------------------- Function for case conversion ----------------------- #


def case_convert(df: pd.DataFrame,
                 cols: Optional[Union[str, List[str]]] = None,
                 to: Optional[str] = 'lower',
                 inplace: Optional[bool] = False) -> Union[pd.DataFrame, None]:
    """
    This helper function converts the cases in the columns of a DataFrame; the aim is to address case inconsistencies 
    in string columns. Use this function to ensure values in a DataFrame are consistent with regards to case, which helps 
    reduce the chance of committing further errors later on in the cleaning process.

    Parameters
    ----------
    df : DataFrame
    cols : str or list of str, optional
        A single column name or a sequence of column names, by default None, which converts all columns that can be inferred as having 'string' dtypes.
    to : str, optional
        The direction or type of case conversion. One of 'lower', 'upper', 'title', or 'capitalize', by default 'lower'.
    inplace : bool, optional
        Whether to return a new DataFrame, by default False.

    Returns
    -------
    DataFrame
        A DataFrame with transformed columns or None if inplace=True.

    Raises
    ------
    TypeError
        The argument 'cols' must be registered as a Sequence or a single string.
    InvalidColumnDtypeError
        User supplied columns contain non-string columns.
    ValueError
        Direction or type of case conversion must either be 'lower', 'upper', 'title', or 'capitalize'.
    """
    if not is_list_str(cols) and cols is not None:
        raise TypeError(
            "'cols' must be a sequence like a list or a single string")

    # Create a copy if inplace=False
    if (not inplace):
        df = df.copy()

    # Branch 1: If user does not specify columns, default to using all columns that are inferred as 'string'
    if cols == None:
        bool_is_str = [is_string_dtype_pd(df[col])
                       for col in df.columns.tolist()]
        cols = list(compress(df.columns.tolist(), bool_is_str))
    # Branch 2: If user does specify, check input
    else:
        # Sub-branch 1: If 'cols' is a single string, then use the string to select column from df directly
        if isinstance(cols, str):
            bool_is_str = [is_string_dtype_pd(df[cols])]
        # Sub-branch 2: If 'cols' is a sequence of strings, then apply list comprehension
        else:
            bool_is_str = [is_string_dtype_pd(df[col]) for col in cols]

        # Take the 'bool_is_str' list from either one of the two sub-branches above, check input column data types
        if not all(bool_is_str):
            raise InvalidColumnDtypeError(col_nms=list(
                compress(cols, [not element for element in bool_is_str])), dtype='string')

    # Use pd.DataFrame since, if user passes a single str as 'cols', df[cols] would return a series, and lambda 'x' would be the elements
    # Apply relies on the Series 'str' attribute to work
    if (to == 'upper'):
        df[cols] = pd.DataFrame(df[cols]).apply(lambda x: x.str.upper())
    elif (to == 'lower'):
        df[cols] = pd.DataFrame(df[cols]).apply(lambda x: x.str.lower())
    elif (to == 'title'):
        df[cols] = pd.DataFrame(df[cols]).apply(lambda x: x.str.title())
    elif (to == 'capitalize'):
        df[cols] = pd.DataFrame(df[cols]).apply(lambda x: x.str.capitalize())
    else:
        raise ValueError(
            "'to' must either by 'lower', 'upper', 'title', or 'capitalize'")

    # Return copy
    if (not inplace):
        return df
    else:
        return None


# ---------------- Function to relocate columns in a DataFrame --------------- #


def relocate(df: pd.DataFrame,
             to_move: Union[str, List[str]],
             before: Union[str, None] = None,
             after: Union[str, None] = None) -> pd.DataFrame:
    """
    This function reorders the columns in a DataFrame based on a reference column, which is either 
    specified as `before` or `after`. Only one of 'before' and 'after' should be supplied as a string.

    Parameters
    ----------
    df : DataFrame
    to_move : str or list of str 
        A single column name or a list of column names to relocate.
    before : str or None, optional
        A reference column before which the columns in 'to_move' should be relocated, by default None.
    after : str or None, optional
        A reference column after which the columns in 'to_move' should be relocated, by default None.

    Returns
    -------
    DataFrame
        A DataFrame with reordered columns.

    Raises
    ------
    TypeError
        The argument 'to_move' must either be a list or a single string.
    TypeError
        Must supply only one of 'before' and 'after' as a string.
    """
    # Check input
    if not is_list_str(to_move):
        raise TypeError(
            "'to_move' must be a sequence like a list or a single string")
    # Exclusive or (exclusive disjunction where this expression is true if and only if the two booleans differ (one is true, the other is false))
    if isinstance(before, str) == isinstance(after, str):
        raise TypeError(
            "must supply only one of 'before' and 'after' as a string")

    # If 'to_move' is a string
    if isinstance(to_move, str):
        # Strings are immutable so this creates a new object that 'to_move' references in the local scope that is mutable
        to_move = [to_move]

    # If the reference column is included in 'to_move', remove it
    if before in to_move:
        to_move.remove(before)
    elif after in to_move:
        to_move.remove(after)

    # List of all column names
    cols = df.columns.tolist()

    # If we wish to move cols before the reference column
    if isinstance(before, str):
        # Select columns before the reference column (not including the reference column)
        seg1 = cols[:cols.index(before)]
        # Segment 2 includes columns to move 'before' the reference column (in other words, append 'before' to the end of 'to_move')
        to_move.append(before)
        seg2 = to_move

    # If we wish to move cols after the reference column
    if isinstance(after, str):
        # Select columns before the reference column (including the reference column)
        seg1 = cols[:cols.index(after) + 1]
        # Segment 2 simply includes columns to move
        seg2 = to_move

    # In either case, 'before' or 'after', we need to ensure columns to move are not in segment 1 (in other words, drop 'to_move' columns from their original positions)
    # Segment 2 in either case should always include columns to move
    seg1 = [col for col in seg1 if col not in seg2]
    # Finally, select the rest of the columns--- those that are not in seg1 and seg2
    seg3 = [col for col in cols if col not in seg1 + seg2]

    # For 'before', seg2 includes columns to move plus the reference column
    # Thus, seg2 created from the 'before' branch ensures that columns to move will appear 'before' the reference column
    # For 'after', adding seg1 (with reference column) + seg2 (to_move) in this order ensures that columns to move will appear 'after' the reference column
    # Finally, seg3 adds the rest of the columns to the end
    return df[seg1 + seg2 + seg3]


# ---------------------------------------------------------------------------- #
#                               Profiling helpers                              #
# ---------------------------------------------------------------------------- #

# ------------------ Function for identifying missing values ----------------- #


def find_missing(df: pd.DataFrame, axis: Optional[int] = 0) -> pd.Series:
    """
    This is a helper function that identifies columns or rows in a DataFrame
    that contain missing values. Then, rows and columns containing values may
    be retrieved using `df.loc[cu.find_missing(df, axis=1).index, :]` or 
    `df.loc[:, cu.find_missing(df, axis=0).index]` respectively.

    Parameters
    ----------
    df : DataFrame
    axis : int, optional
        Whether to identify rows or columns that contain missing values, by default 0 (columns).

    Returns
    -------
    Series of bool
        Boolean series indicating columns or rows with missing values.

    Raises
    ------
    TypeError
        The argument 'axis' must be an integer.
    ValueError
        The argument 'axis' must either be 1 (rows) or 0 (columns).
    """
    if not isinstance(axis, int):
        raise TypeError("The argument 'axis' must be an integer")

    # The lambda function simply returns the true elements of the boolean series
    if axis == 1:
        return df.isna().any(axis=1)[lambda x: x]
    elif axis == 0:
        return df.isna().any(axis=0)[lambda x: x]
    else:
        raise ValueError("'axis' must either be 1 (rows) or 0 (columns)")


# -------------- Function to create a tuple of frequency tables -------------- #


def freq_tbl(df: pd.DataFrame, dropna: Optional[bool] = False, cardinality: Optional[int] = 20, **kwargs: str) -> NamedTuple:
    """
    This function creates a sequence of freqency tables of the fields in a DataFrame, which can be examined to 
    identify misspellings or case inconsistencies. You may pass extra keyword arguments for the underlying pandas 
    function. See `?pandas.DataFrame.value_counts` for options (note that the `subset` argument is not permitted). 
    The function is `dtype` agnostic since the underlying function `pd.Series.value_counts` works on categorical and 
    categorical features.

    Parameters
    ----------
    df : DataFrame
    dropna : bool, optional
        Whether to drop missing values, by default False.
    cardinality : int, optional
        Cardinality limit for the fields in the input DataFrame for filtering out high cardinal fields, by default 20.

    Returns
    -------
    tuple of DataFrame
        A namedtuple of frequency tables containing counts (or proportions if extra keyword arguments are specified) per category for each categorical field.

    Raises
    ------
    ValueError
        Only 'normalize', 'sort', 'ascending' are supported as extra keyword arguments.
    TypeError
        The argument 'cardinality' must be an integer.
    """
    # Check keyword args
    if not all((kwarg in ('normalize', 'sort', 'ascending') for kwarg in kwargs.keys())):
        raise ValueError(
            "Only 'normalize', 'sort', and 'ascending' are supported as extra keyword arguments")
    # Check input
    if not isinstance(cardinality, int):
        raise TypeError("'cardinality' must be an integer")
    # Obtain number of unique values for each column in 'df'
    # Only use columns where the number of unique levels is less than or equal to the cardinality value (defaults to 20)
    df = df.loc[:, [col for col in df.columns if len(
        df[col].unique()) <= cardinality]]
    # Generator of frequency tables
    gen_of_freq = (pd.DataFrame(df[col].value_counts(dropna=dropna, **kwargs))
                   for col in df.columns)
    # Create subclass constructor
    freq = namedtuple('freq', df.columns)
    # Use constructor to create a named tuple
    freq_tbl = freq._make(gen_of_freq)

    return freq_tbl



# ---------------------------------------------------------------------------- #
#                           Load packages and modules                          #
# ---------------------------------------------------------------------------- #

import numpy as np
import pandas as pd

# --------------------------------- Modeling --------------------------------- #

from sklearn.cluster import KMeans
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_array, check_is_fitted

# ------------------------ Elbow method for clustering ----------------------- #

from kneed import KneeLocator
from yellowbrick.cluster.elbow import distortion_score

# --------------------------------- Plotting --------------------------------- #

import matplotlib.pyplot as plt
from matplotlib.collections import PathCollection

# ----------------------------- Standard library ----------------------------- #

from typing import Dict, List, Optional, Union


# ---------------------------------------------------------------------------- #
#                              Custom transformer                              #
# ---------------------------------------------------------------------------- #

class CoordinateTransformer(BaseEstimator, TransformerMixin):
    """
    A custom transformer for handling coordinate data. This transformer creates a column of cluster labels
    using the `sklearn.cluster.KMeans` learning algorithm that may be used in training in lieu of the original 
    coordinate data. Since this transformer accepts only a pandas DataFrame, it can only be used in the `pipeline`
    after an estimator whose `transform` or `fit_transform` returns a pandas DataFrame.


    Parameters
    ----------
    strategy : str, optional
        The strategy for creating clustering labels, by default "kmeans".
    k_range : range, optional
        The range of value for the `n_clusters` parameter in `sklearn.cluster.KMeans` to try in order to determine an optimal value, by default range(4, 13).

    Attributes
    ----------
    distortion_scores_ : List[float]
        A list of mean distortions for each run of `sklearn.cluster.KMeans`. The distortion is computed as the the sum of the squared distances between each observation and 
        its closest centroid. Logically, this is the metric that K-Means attempts to minimize as it is fitting the model.
    optimal_k_: int
        The the optimal number of clusters.
    X_coords: ndarray of shape (n_samples, 2)
        The original coordinate data.
    kmeans_: object
        An instance of `sklearn.cluster.KMeans` fitted with the optimal number of clusters as its `n_clusters` parameter.
    labels_: ndarray of shape (n_samples,)
        The labels of each point.
    """

    def __init__(self, strategy: str = "kmeans", k_range: range = range(4, 13)) -> None:
        self.strategy = strategy
        self.k_range = k_range

    def fit(self, X: pd.DataFrame, y: Optional[pd.Series] = None):
        """
        Fit the transformer on X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            The coordinate (longitude, latitude) columns.
        y : Optional[pd.Series], optional
            Ignored, present here for API consistency by convention, by default None.

        Returns
        -------
        self : object
            A fitted estimator.
        """
        # Input validate, convert coordinate columns to ndarray
        X_coords = check_array(X, accept_sparse=False, dtype="numeric")

        # Multiple runs of k-means
        distortion_scores = []
        for k in self.k_range:
            model = KMeans(n_clusters=k)
            model.fit_transform(X_coords)
            distortion_scores.append(distortion_score(X_coords, model.labels_))

        # Distortion scores
        self.distortion_scores_ = distortion_scores
        # Optimal number of clusters
        self.optimal_k_ = KneeLocator(
            x=self.k_range, y=self.distortion_scores_, curve='convex', direction='decreasing', S=1).knee

        return self

    def transform(self, X: Union[pd.DataFrame, np.ndarray]) -> pd.DataFrame:
        """
        Transform input coordinate columns into a single column of cluster labels.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            The coordinate (longitude, latitude) columns.

        Returns
        -------
         X_transformed : array, shape (n_samples,)
            The labels of each coordinate point.
        """
        # Check that the fit method has been called
        check_is_fitted(self, ('distortion_scores_', 'optimal_k_'))

        # Input validation
        X_coords = check_array(X, accept_sparse=False, dtype="numeric")

        # Coordinate columns
        self.X_coords = X_coords

        # Model
        self.kmeans_ = KMeans(n_clusters=self.optimal_k_, init='k-means++')

        # Store cluster labels as an attribute of the instance
        self.labels_ = self.kmeans_.fit_predict(X_coords)

        return self.labels_

    def plot(self) -> PathCollection:
        """
        Generate a scatter plot of coordinate points with marker colors representing
        the cluster to which each data point belongs. This method should only be called
        once the transformation has completed, i.e., after `transform` or `fit_transform`.

        Parameters
        ----------
        X : pd.DataFrame
            A pandas DataFrame.

        Returns
        -------
        PathCollection
            A collection of Paths, as created by `matplotlib.pyplot.scatter`.

        Raises
        ------
        AttributeError
            The 'labels_' field is created only after `transform` is called on the instance.
        """
        if not hasattr(self, 'labels_'):
            raise AttributeError(
                "This 'CoordinateTransformer' instance is not transformed yet; please call 'transform' with appropriate arguments before plotting")

        # Use numpy.take along columns of the ndarray
        plt.scatter(
            x=np.take(self.X_coords, indices=0, axis=1),
            y=np.take(self.X_coords, indices=1, axis=1),
            c=self.labels_, s=50, cmap='viridis'
        )

        plt.title('K-Means Clustered Map')
        plt.xlabel('Longitude')
        plt.ylabel("Latitude")

        return plt



# ---------------------------------------------------------------------------- #
#                           Load packages and modules                          #
# ---------------------------------------------------------------------------- #

from pandas.api.types import is_string_dtype
from pandas import Series

# ----------------------------- Standard library ----------------------------- #

from collections.abc import Sequence
from typing import List

# ---------------------------------------------------------------------------- #
#                    Helper predicates for input validation                    #
# ---------------------------------------------------------------------------- #

# --------------------------------- Builtins --------------------------------- #


def is_list_str(seq: Sequence) -> bool:
    """
    This helper returns `True` if the input is either a list or a string.

    Parameters
    ----------
    seq : Sequence of objects
        An input sequence to be tested.

    Returns
    -------
    bool
        `True` if the sequence is either a list or a string..
    """
    return isinstance(seq, list) or isinstance(seq, str)

# ----------------------------- Helper for pandas ---------------------------- #


def is_string_dtype_pd(col: Series) -> bool:
    """
    This helper checks whether the provided Series is of the string dtype.

    Parameters
    ----------
    col : Series
        A pandas Series to be tested.

    Returns
    -------
    bool
        `True` if the Series is of the string dtype.
    """
    return is_string_dtype(col)



# ---------------------------------------------------------------------------- #
#                           Load packages and modules                          #
# ---------------------------------------------------------------------------- #

# ----------------------------- Standard library ----------------------------- #

from typing import List

# ---------------------------------------------------------------------------- #
#                          AWS utils module exceptions                         #
# ---------------------------------------------------------------------------- #

# ------------------- Column dtype cannot be inferred error ------------------ #


class ColumnDtypeInferError(Exception):
    """
    Exception raised when a data frame contains column(s) with dtype(s) that cannot be inferred.
    """

    def __init__(self, col_nms: List[str]) -> None:
        self.col_nms = col_nms

    def __str__(self) -> str:
        return f'The dtypes of the following columns cannot be inferred: {self.col_nms}'

# ---------------------------------------------------------------------------- #
#                        Data cleaning module exceptions                       #
# ---------------------------------------------------------------------------- #

# ------------------------------- Keyword error ------------------------------ #


class ColumnNameKeyWordError(Exception):
    """
    Exception raised when column names are keywords.
    """

    def __init__(self, col_nms: List[str]) -> None:
        self.col_nms = col_nms

    def __str__(self) -> str:
        return f'Columns {self.col_nms} are keywords of the language, and cannot be used as ordinary identifiers'

# -------------------------- Start with digit error -------------------------- #


class ColumnNameStartWithDigitError(Exception):
    """
    Exception raised when column names start with digits.
    """

    def __init__(self, col_nms: List[str]) -> None:
        self.col_nms = col_nms

    def __str__(self) -> str:
        return f'Columns {self.col_nms} must not start with digits'

# ------------------------ Catch-all identifier error ------------------------ #


class InvalidIdentifierError(Exception):
    """
    Exception raised when column names are invalid identifiers.
    """

    def __init__(self, col_nms: List[str]) -> None:
        self.col_nms = col_nms

    def __str__(self) -> str:
        return f'Columns {self.col_nms} are invalid identifiers'

# ----------------- Columns do not have correct dtypes error ----------------- #


class InvalidColumnDtypeError(Exception):
    """
    Exception raised when user supplied columns do not have the dtypes that the function expects.
    """

    def __init__(self, col_nms: List[str], dtype: str) -> None:
        self.col_nms = col_nms
        self.dtype = dtype

    def __str__(self) -> str:
        return f"Columns {self.col_nms} are invalid as dtype '{self.dtype}' is expected"



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['my_mltools']

package_data = \
{'': ['*']}

install_requires = \
['kneed>=0.7.0,<0.8.0',
 'pandas>=1.4.2,<2.0.0',
 'scikit-learn>=1.0.2,<2.0.0',
 'tensorflow>=2.8.0,<3.0.0',
 'yellowbrick>=1.4,<2.0']

setup_kwargs = {
    'name': 'my-mltools',
    'version': '0.5.0',
    'description': 'My machine learning toolkit.',
    'long_description': '# my_mltools\n\n[![codecov](https://codecov.io/gh/YangWu1227/my_mltools/branch/main/graph/badge.svg?token=9ZL267TMVD)](https://codecov.io/gh/YangWu1227/my_mltools)\n\nMy machine learning toolkit.\n\n## Installation\n\n```bash\n$ pip install my_mltools\n```\n## License\n\n`my_mltools` was created by Yang Wu. It is licensed under the terms of the MIT license.\n\n## Credits\n\n`my_mltools` was created with [`cookiecutter`](https://cookiecutter.readthedocs.io/en/latest/) and the `py-pkgs-cookiecutter` [template](https://github.com/py-pkgs/py-pkgs-cookiecutter).\n',
    'author': 'Yang Wu',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
